/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "dbacl.h"

/* look for these where main() is defined */

/* these globals will have to be refactored once
   I decide on an API for libdbacl ... */

extern options_t options; 

extern Empirical empirical;

extern Category cat[MAX_CAT];
extern category_count_t cat_count;

extern Regex re[MAX_RE];
extern regex_count_t regex_count;

int hashfull_warning;

/***********************************************************
 * MISCELLANEOUS FUNCTIONS                                 *
 ***********************************************************/

char *sanitize_path(char *in) {
  char *q;
  char *path;
  charbuf_len_t l;

  /* this bit likely fails in DOS ;-) */
  if( (*in != '/') && (*in != '.') && (path = getenv("DBACL_PATH")) ) {
    l = strlen(path);
    q = malloc(l + strlen(in) + 3);
    strcpy(q, path);
    if( q[l - 1] != '/' ) {
      q[l] = '/'; 
      q[l + 1] = 0;
    }
    strcat(q, in);
    return q;
  } else {
    return in;
  }
}

digitized_weight_t digitize_a_weight(weight_t w) {
  if( w < -320 ) {
    return DIGITIZED_WEIGHT_MIN;
  } else if( 320 < w ) {
    return DIGITIZED_WEIGHT_MAX;
  } else {
    return 100 * w;
  }
}

/***********************************************************
 * EMPIRICAL DISTRIBUTION OF TOKENS                        *
 ***********************************************************/

/* initialize global learner object */
void init_empirical(Empirical *emp, hash_count_t dmt, hash_bit_count_t dmhb) {

    /* some constants */
    emp->max_tokens = dmt;
    emp->max_hash_bits = dmhb;
    emp->full_token_count = 0;
    emp->unique_token_count = 0;
    emp->track_features = 0;
    emp->feature_stack_top = 0;

    /* allocate room for hash */
    emp->hash = calloc(emp->max_tokens, sizeof(h_item));
    if( !emp->hash ) {
	fprintf(stderr, 
		"error: not enough memory? I couldn't allocate %li bytes\n", 
		(sizeof(h_item) * ((long int)emp->max_tokens)));
	exit(0);
    }
}

void clear_empirical(Empirical *emp) {
    token_stack_t i;

    if( emp->track_features ) {
	/* this may actually be slower than a global memset */ 
	for(i = 0; i < emp->feature_stack_top; i++) {
	    memset(emp->feature_stack[i], 0, sizeof(h_item));
	}
    } else {
	memset(emp->hash, 0, sizeof(h_item) * emp->max_tokens);
    }

    emp->full_token_count = 0;
    emp->unique_token_count = 0;
    emp->feature_stack_top = 0;

    if( options & (1<<OPTION_FASTEMP) ) {
      emp->track_features = 1;
    }
}

h_item *find_in_empirical(Empirical *emp, hash_value_t id) {
  register h_item *i, *loop;
  /* start at id */
  i = loop = &emp->hash[id & (empirical.max_tokens - 1)];

  while( FILLEDP(i) ) {
    if( EQUALP(i->id,id) ) {
      return i; /* found id */
    } else {
      i++; /* not found */
      /* wrap around */
      i = (i >= &emp->hash[emp->max_tokens]) ? emp->hash : i; 
      if( i == loop ) {
	return NULL; /* when hash table is full */
      }
    }
  }

  /* empty slot, so not found */

  return i; 
}

/* calculates the entropy of the empirical measure */
score_t empirical_entropy(Empirical *emp) {
    hash_count_t i;
    score_t e = 0.0;

    if( emp->track_features && 
	(emp->feature_stack_top < MAX_TOKEN_LINE_STACK) ) {
	for(i = 0; i < emp->feature_stack_top; i++) {
	    e += ((score_t)emp->feature_stack[i]->count) * 
	      log((score_t)emp->feature_stack[i]->count);
	}
    } else {
	for(i = 0; i < emp->max_tokens; i++) {
	    if( FILLEDP(&emp->hash[i]) ) {
		e += ((score_t)emp->hash[i].count) * 
		  log((score_t)emp->hash[i].count);
	    }
	}
    }
    e = e/emp->full_token_count - log((score_t)emp->full_token_count);

    return -e;
}

/***********************************************************
 * CATEGORY FUNCTIONS                                      *
 ***********************************************************/

/* initialize to zero. Filename specified elsewhere */
void init_category(Category *cat) {
    char *p;

    cat->model_full_token_count = 0;
    cat->model_unique_token_count = 0;
    cat->score = 0.0;
    cat->complexity = 0;
    cat->max_order = 0;
    if( (p = strrchr(cat->filename, '/')) ) {
	cat->filename = p + 1; /* only keep basename */
    }
    cat->retype = 0;
    cat->model_type = simple;

    cat->hash = NULL;
}

/* frees the resrouces associated with a category */
void free_category(Category *cat) {
  if( cat->hash ) {
    free(cat->hash);
  }
}

/* turns purely random text into a category of its own */
void init_purely_random_text_category(Category *cat) {
  alphabet_size_t i, j;
  weight_t z = -log((double)ASIZE - 1.0);

#if defined DIGITIZE_DIGRAMS
  digitized_weight_t zz = PACK_DIGRAMS(z);
#endif
    
  for(i = 1; i < ASIZE; i++) {
    for(j = 1; j < ASIZE; j++) {
#if defined DIGITIZE_DIGRAMS
      cat->dig[i][j] = zz;
#else
      cat->dig[i][j] = z;
#endif
    }
  }

  z = -log((double)ASIZE - 2.0);
#if defined DIGITIZE_DIGRAMS
  zz = PACK_DIGRAMS(z);
#endif
  for(j = 1; j < ASIZE; j++) {
#if defined DIGITIZE_DIGRAMS
    cat->dig[DIAMOND][j] = zz;
#else
    cat->dig[DIAMOND][j] = z;
#endif
  } 

  /* not needed: set DIAMOND-DIAMOND score for completeness only */

#if defined DIGITIZE_DIGRAMS
  cat->dig[DIAMOND][DIAMOND] = DIGITIZED_WEIGHT_MIN;
#else
  cat->dig[DIAMOND][DIAMOND] = log(0.0);
#endif

  cat->logZ = 0.0;
  cat->hash = NULL;
  cat->model_type = simple;
  cat->max_order = 1;
  options |= (1<<OPTION_NOREGEX); 
}

c_item *find_in_category(Category *cat, hash_value_t id) {
    register c_item *i, *loop;

    if( cat->hash ) {
	/* start at id */
	i = loop = &cat->hash[id & (cat->max_tokens - 1)];

	while( FILLEDP(i) ) {
	    if( EQUALP(i->id,id) ) {
		return i; /* found id */
	    } else {
		i++; /* not found */
		/* wrap around */
		i = (i >= &cat->hash[cat->max_tokens]) ? cat->hash : i; 
		if( i == loop ) {
		    return NULL; /* when hash table is full */
		}
	    }
	}
	return i;
    } else {
	return NULL;
    }
}

/* for each loaded category, this calculates the score */
void score_word(char *tok, token_order_t r, regex_count_t re) {
  category_count_t i;
  alphabet_size_t pp, pc;
  hash_value_t id;
  char *q;
  register c_item *k;
  h_item *h = NULL;

  if( *(tok + 2) ) { /* DIAMOND-DIAMOND is empty string */

    id = (hash_value_t)hash(tok, strlen(tok), 0);

    if( (options & (1<<OPTION_CALCENTROPY)) && (r == 1) ) {
      /* add the token to the hash */

      h = find_in_empirical(&empirical, id);
      if( h ) {
 	if( FILLEDP(h) ) {
	  h->count += ( h->count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
 	} else {
 	  if( /* !FILLEDP(i) && */
 	     ((100 * empirical.unique_token_count) <
 	      (HASH_FULL * empirical.max_tokens) )) {
 	    /* fill the empirical hash */
 	    SET(h->id,id);
	    empirical.unique_token_count += 
	      ( empirical.unique_token_count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
	    h->count += ( h->count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
 	  } else {
 	    /* hash full */
 	    h = NULL; 
 	    if( !hashfull_warning ) {
 	      fprintf(stderr,
 		      "warning: empirical hash full, calculation may be skewed. " 
 		      "Try option -h %d\n", 
 		      (empirical.max_hash_bits + 1));
 	      hashfull_warning = 1;
 	    }
 	    return; /* pretend word doesn't exist */
 	  }

	  if( empirical.track_features ) {
	    if( empirical.feature_stack_top < MAX_TOKEN_LINE_STACK ) {
	      empirical.feature_stack[empirical.feature_stack_top++] = h;
	    } else {
	      empirical.track_features = 0;
	      empirical.feature_stack_top = 0;
	    }
	  }
 	}
	empirical.full_token_count += 
	  ( empirical.full_token_count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
	
      }
    }

    /* now do scoring for all available categories */
    for(i = 0; i < cat_count; i++) {

      switch(cat[i].model_type) {
      case simple:
	/* see if this is for us */
	if( ((re == 0) && (r == 1)) ||
	    ((re > 0) && (cat[i].retype & (1<<(re-1)))) ) {
	  /* if token found, add its lambda weight */
	  k = find_in_category(&cat[i], id);
	  if( k ) {
	    cat[i].score += UNPACK_LAMBDA(k->lam);
	  }

	  /* now compute the reference weight from digram model */
	  pp = (unsigned char)*tok;
	  q = tok + 1;
	  while( *q ) {
	    if( *q == '\r' ) {
	      q++;
	      continue;
	    }
	    pc = (unsigned char)*q;
	    cat[i].score += UNPACK_DIGRAMS(cat[i].dig[pp][pc]);
	    pp = pc;
	    q++;
	  }

	  /* don't forget the normalizing constant */
	  cat[i].score -= cat[i].logZ;      

	  /* update complexity for this category */
	  cat[i].complexity++;

	  /* add correction factor for multinomial case */
	  if(h) {
	    cat[i].score += log((weight_t)cat[i].complexity) - 
	      log((weight_t)h->count);
	  }
	}
	break;

      case sequential:
      default:

	/* see if this is for us */
	if( ((re == 0) && !cat[i].retype) ||
	    ((re > 0) && (cat[i].retype & (1<<(re-1)))) ) {

	  /* if token found, add its lambda weight */
	  k = find_in_category(&cat[i], id);
	  if( k ) {
	    cat[i].score += UNPACK_LAMBDA(k->lam);
	  }

	  if( r == 1 ) {
	    /* now compute the reference weight from digram model */
	    pp = (unsigned char)*tok;
	    q = tok + 1;
	    while( *q ) {
	      if( *q == '\r' ) {
		q++;
		continue;
	      }
	      pc = (unsigned char)*q;
	      cat[i].score += UNPACK_DIGRAMS(cat[i].dig[pp][pc]);
	      pp = pc;
	      q++;
	    }
	  }

	  if( r == cat[i].max_order ) {
	    /* don't forget the normalizing constant */
	    cat[i].score -= cat[i].logZ;      

	    /* update complexity for this category */
	    cat[i].complexity++;
	  }
	}
	break;

      }

    }
  }
}

/***********************************************************
 * FILE MANAGEMENT FUNCTIONS                               *
 ***********************************************************/

/* loads a category hash */
error_code_t load_category(Category *cat) {
  hash_count_t i, j;
  regex_count_t c;
  char buf[MAGIC_BUFSIZE];
  char scratchbuf[MAGIC_BUFSIZE];
  char *p;

  FILE *input;

#if defined HAVE_LIBBOOST_REGEX
  charbuf_len_t r;
#endif

  if( (input = fopen(cat->filename, "rb")) ) {
    fgets(buf, MAGIC_BUFSIZE, input);
    if( strncmp(buf, MAGIC1, 16 + strlen(VERSION)) ) {
      fprintf(stderr, 
	      "error: not a dbacl version "VERSION" category file [%s]\n", 
	      cat->filename);
      return 0;
    } 

    init_category(cat); /* changes filename */

    fgets(buf, MAGIC_BUFSIZE, input);
    sscanf(buf, MAGIC2_i, &cat->divergence, &cat->logZ, 
	   &cat->max_order, scratchbuf);
    if( scratchbuf[0] == 'm' ) {
      cat->model_type = simple;
    } else {
      cat->model_type = sequential;
    }
    fgets(buf, MAGIC_BUFSIZE, input);
    sscanf(buf, MAGIC3, 
	   &cat->max_hash_bits, 
	   &cat->model_full_token_count, 
	   &cat->model_unique_token_count,
	   &cat->model_num_docs);
    cat->max_tokens = (1<<cat->max_hash_bits);

    /* see if there are any regexes */
    fgets(buf, MAGIC_BUFSIZE, input);
    while(1) {
      if( strncmp(buf, MAGIC6, 2) == 0 ) {
	break;
      } else if( strncmp(buf, MAGIC5_i, 8) == 0 ) {

	/* set up the submatch bitmap */
	re[regex_count].submatches |= 0;

	if( (p = strrchr(buf + RESTARTPOS, '|')) && ( *(--p) == '|') ) {
	  /* assume string ends in ||12345, use as bitmap */
	  *p = '\0';
	  for(p += 2; *p; p++) {
	    /* assume ascii number positions */
	    if( (*p > '9') || (*p < '1')) {
	      if( *p != '\n' ) {
		fprintf(stderr, 
			"warning: could not decode || suffix for %s while loading %s\n", 
			buf + RESTARTPOS, cat->filename);
	      }
	    } else {
	      re[regex_count].submatches |= (1<<(*p - '0'));
	    }
	  }
	} else { /* no bitmap specified */
	  re[regex_count].submatches = ~0;
	}

#if defined HAVE_LIBBOOST_REGEX
	/* boost regexes accept wide characters */
	    
	r = strlen(buf + RESTARTPOS);
	re[regex_count].string = malloc((r + 1) * sizeof(wchar_t));
	/* note: skip the trailing newline */
	r = mbstowcs(re[regex_count].string, buf + RESTARTPOS, r - 1); 
	if( r <= 0 ) {
	  fprintf(stderr, 
		  "error: couldn't convert regular expression '%s'.\n",
		  buf + RESTARTPOS);
	  exit(0); /* no point in going on */
	} else {
	  re[regex_count].string[r] = L'\0';
	}
	/* now see if we already have this string */
	for(c = 0; c < regex_count; c++) {
	  if( wcscmp(re[c].string, re[regex_count].string) == 0 ) {
	    break;
	  }
	}
	if( c < regex_count ) {
	  /* we found this regex, no need to repeat it */
	  free(re[regex_count].string);
	} else {
	  /* compile the regex (uses regcompW) */
	  if( regcomp(&re[regex_count].regex,
		      re[regex_count].string, REG_EXTENDED) != 0 ) {
	    fprintf(stderr, 
		    "error: could not compile regular expression '%ls' for %s.\n", 
		    re[regex_count].string, cat->filename);
	    exit(0);
	  } else {
	    regex_count++;
	    if( regex_count >= MAX_RE ) { 
	      fprintf(stderr, "error: too many regular expressions\n");
	      exit(0); /* no point in going on */
	    }
	  }
	}
#else
	/* remove trailing newline */
	if( buf[strlen(buf) - 1] == '\n' ) { 
	  buf[strlen(buf) - 1] = '\0'; 
	}
	/* GNU regexes use ordinary strings */
	/* it's a regex - see if we've got it already */
	for(c = 0; c < regex_count; c++) {
	  if( strcmp(re[c].string, buf + RESTARTPOS) == 0 ) {
	    break;
	  }
	}
	if( c >= regex_count ) { /* not found */
	  /* add it to our list */
	  re[regex_count].string = strdup(buf + RESTARTPOS);
	  /* and compile the regex */
	  if( regcomp(&re[regex_count].regex, 
		      re[regex_count].string, REG_EXTENDED) != 0 ) {
	    fprintf(stderr, 
		    "error: could not compile regular expression '%s' for %s.\n", 
		    re[regex_count].string, cat->filename);
	    exit(0); /* not much point going on */
	  } else {
	    regex_count++;
	    if( regex_count >= MAX_RE ) { 
	      fprintf(stderr, "error: too many regular expressions\n");
	      exit(0); /* no point in going on */
	    }
	  }
	}
#endif
	/* now flag the regex as taken */
	cat->retype |= (1<<c);

      } else if( strncmp(buf, MAGIC4, 10) == 0) {
	/* it's the case sensitive flag */
	options |= (1<<OPTION_CASEN);
      } else if( strncmp(buf, MAGIC7, 5) == 0) {
	options |= (1<<OPTION_I18N);
      }

      /* finished with current line, get next one */
      fgets(buf, MAGIC_BUFSIZE, input); 
    }

    /* if this category did not register a regex, it wants
       the default processing, so we flag this */
    if( !cat->retype ) {
      options |= (1<<OPTION_NOREGEX);
    }

    /* read character frequencies */
    fread(cat->dig, SIZEOF_DIGRAMS, ASIZE * ASIZE, input);

    /* allocate hash table */
    if( !(cat->hash = malloc(sizeof(c_item) * cat->max_tokens)) ) {
      fprintf(stderr, "error: not enough memory for category %s\n", 
	      cat->filename);
      return 0;
    }
    /* read in hash table */
    i = cat->max_tokens;
    j = 0;
    while( j < i ) {
      j += fread(cat->hash + j, sizeof(c_item), i, input);
    }
    fclose(input);
  } else {
    fprintf(stderr, "error: cannot open file for reading %s\n", 
	    cat->filename);
    return 0;
  }

  return 1;
}
